/* Dynamic classes
  
   Bob Gray
   Gisli Hjalmtysson
   Network Mathematics Research Department (AK0112730), AT&T 
   600 Mountain Avenue
   Murray Hill, New Jersey

   4 June 1996

   faststring.h

   This file defines fast versions of selected string commands.
*/

#ifndef _FAST_STRING_H
#define _FAST_STRING_H

inline int fast_strlen (const char *string)
{
    register int len = 0;

    while (*string++ != 0) {
	len++;
    }

    return (len);
}

inline char *fast_strcpy (char *dest, const char *src)
{
    while (*src != '\0') {
	*dest++ = *src++;
    }

    *dest = '\0';
    return (dest);
}

inline char *fast_atoi (int n, char *buffer)
{
    register char c;
    register char *sp = buffer;
    register char *dp = buffer;

	/* construct the string in reverse order */

    do {
	*sp++ = (n % 10) + '0';
	n /= 10;
    } while (n > 0);

	/* reverse the string */

    *sp-- = '\0';

    while (sp > dp) {
	c = *dp;
	*dp++ = *sp;
	*sp-- = c;
    }

    return (buffer);
}

const unsigned char lookup_lower[256] = 
{ 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,      /* 00 */
  0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,      /* 08 */
  0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,      /* 10 */
  0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,      /* 18 */
  0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,      /* 20 */
  0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,      /* 28 */
  0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,      /* 30 */
  0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,      /* 38 */
  0x40, 'a' , 'b' , 'c' , 'd' , 'e' , 'f' , 'g' ,      /* 40 */
  'h' , 'i' , 'j' , 'k' , 'l' , 'm' , 'n' , 'o' ,      /* 48 */
  'p' , 'q' , 'r' , 's' , 't' , 'u' , 'v' , 'w' ,      /* 50 */
  'x' , 'y' , 'z' , 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,      /* 58 */
  0x60, 'a' , 'b' , 'c' , 'd' , 'e' , 'f' , 'g' ,      /* 60 */
  'h' , 'i' , 'j' , 'k' , 'l' , 'm' , 'n' , 'o' ,      /* 68 */
  'p' , 'q' , 'r' , 's' , 't' , 'u' , 'v' , 'w' ,      /* 70 */
  'x' , 'y' , 'z' , 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,      /* 78 */
  0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,      /* 80 */
  0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,      /* 88 */
  0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,      /* 90 */
  0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,      /* 98 */
  0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7,      /* A0 */
  0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF,      /* A8 */
  0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7,      /* B0 */
  0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF,      /* B8 */
  0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7,      /* C0 */
  0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,      /* C8 */
  0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7,      /* D0 */
  0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,      /* D8 */
  0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7,      /* E0 */
  0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,      /* E8 */
  0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7,      /* F0 */
  0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF       /* F8 */
}; 
   
inline char fast_tolower (char ch) 
{
    return (lookup_lower[ch]);
}

    /* two simplified, fast memcpy routines */

inline char *fast_memcpy (char *dp, const char *sp, unsigned size)
{
  register unsigned i;

  for (i = 0; i < size; i++) {
    *dp++ = *sp++;
  }

  return (dp);
}

inline char *fast_src_memcpy (char *dp, char *sp, unsigned size)
{
  register unsigned i;

  for (i = 0; i < size; i++) {
    *dp++ = *sp++;
  }

  return (sp);
}

#endif
