/*
 * Mgraphics.c
 *		Motif/Lesstif BGI-like grahphic library module
 *
 * Programmer: Park Jae-Ho(jrogue@netian.com), Lee Hai-Young
 *
 * This project is inspired by the WINIO library in UNDOCUMENTED WINDOWS.
 * This library can be copied, modified, used by others,
 * BUT PLEASE PROVIDE THAT ABOVE NOTICE APPEAR IN ALL COPIES OF THIS LIBRARY.
 */

#include <Xm/Xm.h>
#include <Xm/MainW.h> 
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <stdio.h>
#include "Mgraphics.h"
#include "Wcommon.h"

/* necessary Widgets */

static Widget graphPop ;
static Widget graphMain ;
static Widget graphForm ;	/* moved by rogue for template consistency */

/* necessary global variables */

MScreen *curScreen ;
Colormap curColormap ;
char *BasicColor[16] = {
	"black", 
	"blue",
	"green",
	"cyan",
	"red",
	"magenta",
	"brown",
	"lightgray",
	"darkgray",
	"lightblue",
	"lightgreen",	/* inserted by rogue for color completeness */
	"lightcyan",
	"lightred",
	"lightmagenta",
	"yellow",
	"white" 
} ; 

/* definition of private callback function */
/* changed by rogue for simplicity */
void exposeCB ( Widget w, caddr_t clientData, caddr_t userData );


void MinitGraph ( int argc, char **argv )
{
	int n ;
	Arg wargs[10] ;

	WcommonInit( argc, argv ) ;

	graphPop = XtCreatePopupShell( "GRAPH", applicationShellWidgetClass,
                                        topLevel, NULL, 0 ) ;
	graphMain = XtVaCreateManagedWidget( "graphMain",
                        xmMainWindowWidgetClass, graphPop,
                        XmNvisualPolicy, XmVARIABLE,
			XmNshowSeparator, True,
                        NULL, 0 ) ;

	/* check for several form widget.... */	
	/* moved by rogue for initialize */
	graphForm = XmCreateForm( graphMain, "graphForm", NULL, 0 ) ; 
	XtManageChild( graphForm );

	XmMainWindowSetAreas( graphMain, NULL, NULL, NULL, NULL,
			graphForm ) ;
	/* end of modifications */

	curColormap = DefaultColormap( display, DefaultScreen( display ) ) ;
}


void MloopGraph ( void )
{
        XtMainLoop();
}

void showGraph ( void )
{
        XtPopup( graphPop, XtGrabNone );
        XMapRaised( display, XtWindow( graphPop ) );
	exposeCB( curScreen->screen, curScreen, NULL );
	XFlush( display );
}

void hideGraph ( void )
{
        XUnmapWindow( display, XtWindow( graphPop ) );
	XFlush( display );
}

/* changed by rogue for being able to set width & hegith of screen */
MScreen *CreateScreen ( int iwidth, int iheight )
{
	MScreen *temp ;
	int 	n ;
	Arg	wargs[10] ;
	
	temp = (MScreen *)malloc(sizeof(MScreen)) ;

	/* inserted by rogue for compatiblity */	
	n = 0 ;
        XtSetArg( wargs[n], XmNwidth,           iwidth ) ;     n++ ;
        XtSetArg( wargs[n], XmNheight,          iheight ) ;    n++ ;
        XtSetArg( wargs[n], XmNresizePolicy,    XmRESIZE_NONE ) ;n++ ;

	temp->width = iwidth;
	temp->height = iheight;
	/* end of modifications */

	temp->screen = XmCreateDrawingArea( graphForm, "graphWork", wargs, n ) ;

	/* inserted by rogue for curing Sonne's mistake */
	XtManageChild( temp->screen );

	XtAddCallback( temp->screen, XmNexposeCallback, exposeCB, temp ) ;

	temp->pixmap = XCreatePixmap(display, DefaultRootWindow( display )
			,iwidth, iheight, DefaultDepth( display, 0 ) ) ;

	temp->gc = XCreateGC( display, temp->pixmap, 0, 0) ;
	
	/* inserted by rogue for initialize garbage pixmap memory */
	/* forground : white, background : black */
	XSetForeground( display, temp->gc, BlackPixel( display,
			DefaultScreen( display ) ) );
	XSetBackground( display, temp->gc, WhitePixel( display, 
			DefaultScreen( display ) ) );
	XFillRectangle( display, temp->pixmap, temp->gc, 0, 0,
				temp->width, temp->height );
	/* end of modifications */

	temp->curix = temp->curiy = 0 ;
	temp->next = NULL ;

	/* insert current screen list */

	if ( curScreen == NULL ) {
		curScreen = temp ;
		temp->next = temp ;
		return temp ;
	}
	else {	
		temp->next = curScreen->next ;
		curScreen->next = temp ;
		curScreen = temp ;
		return temp ;
	}
} 

void SetScreen ( target )
MScreen *target ; 
{
	MScreen *temp ;

	temp = curScreen ;
	while ( temp != target ) {
		if ( temp->next == curScreen ) {
			printf("\nerror : undefined Screen. ") ;
			return ;
		}
		temp = temp->next ;
	}
	
	curScreen = temp ;
} 

void clearScreen ( void )
{
	WclearScreen( curScreen ) ;
}	

int  getmaxx ( void )
{
	return( Wgetmaxx( curScreen ) ) ;
}

int  getmaxy ( void )
{
	return ( Wgetmaxy( curScreen ) ) ;
}

void putpixel ( ix, iy )
int  ix, iy ;
{
        Wputpixel( curScreen, ix, iy ) ;
}

void line ( ixStart, iyStart, ixFinish, iyFinish )
int  ixStart, iyStart, ixFinish, iyFinish ;
{
	Wline( curScreen, ixStart, iyStart, ixFinish, iyFinish ) ;
}

void setcolor ( colorno )
int  colorno ;
{
	Wsetnamedcolor( curScreen, BasicColor[colorno] ) ;
}

void setnamedcolor ( colorname )
char *colorname ;
{
	Wsetnamedcolor( curScreen, colorname ) ;
}

void setbkcolor ( colorno )
int  colorno ;
{
	Wsetnamedbkcolor( curScreen, BasicColor[colorno] ) ;
}

void setnamedbkcolor ( colorname )
char *colorname ;
{
	Wsetnamedbkcolor( curScreen, colorname ) ;
}

/* inserted by rogue for reading a key from screen */
char Ggetch ( void )
{
	return ( Wgetch( curScreen->screen ) );
}

void exposeCB ( Widget w, caddr_t clientData, caddr_t userData )
{
	MScreen *scrT = clientData;
	/* changed by rogue for simplicity */
	XCopyArea( display, scrT->pixmap, XtWindow( scrT->screen ),
		scrT->gc, 0, 0, scrT->width, scrT->height, 0, 0 ) ;
}
