/* Copyright (c) 1992 by AT&T Bell Laboratories. */
/* Advanced C++ Programming Styles and Idioms */
/* James O. Coplien */
/* All rights reserved. */

/*-----------------------------------------------------------*/
/*                                                           */
/* File bshapes.cc,     ڵ带         */
/*                                                           */
/*-----------------------------------------------------------*/

#include "bshapes.h"
extern "C" {
#include "Mgraphics.h"
}

/*
 * CLASS ANGLE
 */

Angle
Angle::operator+(const Angle& angle) const {
    Angle retval = angle;
    retval.radians += radians;
    return retval;
}

/*
 * CLASS COORDINATE
 */

void
Coordinate::rotateAbout(Coordinate pivot, Angle angle)
{
/* ߸ ǥ  ˰̹Ƿ ּ ó
    long xdistance = pivot.x()-x(), ydistance = pivot.y()-y();
    long xdistsquared = xdistance * xdistance,
    	ydistsquared = ydistance * ydistance;
    double r = ::sqrt( xdistsquared + ydistsquared );
    Angle newangle = angle +
                Angleatan(double(ydistance)/double(xdistance));
    x_rep = pivot.x() + long(r * ::cos(newangle));
    y_rep = pivot.y() + long(r * ::sin(newangle));
*/

/*  ǥ  ˰ */
    x_rep = long((y() - pivot.y()) * sin(angle) + (x() - pivot.x()) * cos(angle) + pivot.x());
    y_rep = long((y() - pivot.y()) * cos(angle) - (x() - pivot.x()) * sin(angle) + pivot.y());
}

/*
 * CLASS SHAPE
 */

/* Ϻ ׷ Ű  ÷ */

Shape::Shape() {
    center_val = Coordinate(0,0);
    next_pointer=prev_pointer=0;
    color = White;
}

Shape::~Shape() {
}

/*
 * CLASS LINE
 */

void
Line::rotateAbout(Coordinate c, Angle angle) {
    erase();
    p.rotateAbout(c, angle);
    q.rotateAbout(c, angle);
    draw();
}

void
Line::draw() {
    if (p.x()-q.x() || p.y()-q.y()) {
        setcolor(WHITE);
        line(p.x(), p.y(), q.x(), q.y());
    }
}

Line::Line(Coordinate p1, Coordinate p2) {
    p = p1; q = p2;
    center_val = Coordinate((p.x()+q.x())/2, (p.y()+q.y())/2);
    color = Color(White);
    draw();
}

/*
 * CLASS RECTANGLE
 */

void
Rect::rotate(Angle angle) {
    erase();
    l1.rotateAbout(center(), angle);
    l2.rotateAbout(center(), angle);
    l3.rotateAbout(center(), angle);
    l4.rotateAbout(center(), angle);
    draw();
}

void
Rect::draw() {
    l1.draw();
    l2.draw();
    l3.draw();
    l4.draw();
}

void
Rect::move(Coordinate c) {
    /*
     * Ű ̴߽.
     * ̸ ߽  .  ũ⸸ŭ   ߽ 
     *  󸶸ŭ ִ   ġŲ.
     */

    erase();
    long xmoved = c.x() - center().x();
    long ymoved = c.y() - center().y();
    center_val = c;

    l1 = Line(Coordinate(l1.origin().x()+xmoved,
                                      l1.origin().y()+ymoved),
    	    Coordinate(l1.e().x()+xmoved,l1.e().y()+ymoved) );
    l2 = Line(Coordinate(l2.origin().x()+xmoved,
                                      l2.origin().y()+ymoved),
    	    Coordinate(l2.e().x()+xmoved,l2.e().y()+ymoved) );
    l3 = Line(Coordinate(l3.origin().x()+xmoved,
                                      l3.origin().y()+ymoved),
    	    Coordinate(l3.e().x()+xmoved,l3.e().y()+ymoved) );
    l4 = Line(Coordinate(l4.origin().x()+xmoved,
                                      l4.origin().y()+ymoved),
    	    Coordinate(l4.e().x()+xmoved,l4.e().y()+ymoved) );
    draw();
}

Rect::Rect(Coordinate topLeft, long xsize, long ysize) {
    Coordinate    a(topLeft);
    Coordinate    b(a.x()+xsize, a.y());
    Coordinate    c(a.x()+xsize, a.y()+ysize);
// ε 簢 ߽   d ߴ.
    Coordinate    d(a.x(),a.y()+ysize);
// ε 簢 ׸  l1, l2, l3, l4  ٲ.
    l1 = Line(a,b);
    l2 = Line(b,c);
    l3 = Line(c,d);
    l4 = Line(d,a);
    center_val = Coordinate((l1.origin().x()+l2.e().x())/2,
    	    	    (l1.origin().y()+l2.e().y())/2);
    draw();
}

/*
 * CLASS TRIANGLE
 */

void
Triangle::rotate(Angle angle) {
    erase();
    l1.rotateAbout(center(), angle);
    l2.rotateAbout(center(), angle);
    l3.rotateAbout(center(), angle);
    draw();
}

void
Triangle::move(Coordinate c) {
    /*
     * Ű ̴߽.
     * ̸ ߽  .  ũ⸸ŭ   ߽ 
     *  󸶸ŭ ִ   ġŲ.
     */

    erase();
    long xmoved = c.x() - center().x();
    long ymoved = c.y() - center().y();
    center_val = c;
    l1 = Line(Coordinate(l1.origin().x()+xmoved,
                                    l1.origin().y()+ymoved),
    	    Coordinate(l1.e().x()+xmoved,l1.e().y()+ymoved) );
    l2 = Line(Coordinate(l2.origin().x()+xmoved,
                                    l2.origin().y()+ymoved),
    	    Coordinate(l2.e().x()+xmoved,l2.e().y()+ymoved) );
    l3 = Line(Coordinate(l3.origin().x()+xmoved,
                                    l3.origin().y()+ymoved),
    	    Coordinate(l3.e().x()+xmoved,l3.e().y()+ymoved) );
    draw();
}

void
Triangle::draw() {
    l1.draw(); l2.draw(); l3.draw();
}

Triangle::Triangle(Coordinate a, Coordinate b, Coordinate c) {
    l1 = Line(a,b);  l2 = Line(b,c);  l3 = Line(c,a);
    center_val =
               Coordinate((l1.e().x()+l2.e().x()+l3.e().x())/3,
        	    	    (l1.e().y()+l2.e().y()+l3.e().y())/3);
    draw();
}

