/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 6-5 */

// comparing the performance of two associative arrays

#include <iostream>
#include <string>
#include <map>
#include <string.h>
#include <time.h>

using namespace std;

const char *COMMAND_SIZE="!size";  
const char *COMMAND_PRINT="!print";

class AssociativeArrayMap {
public:
    AssociativeArrayMap() { }

    void  atPut(void* element, const char* key) {
        if (!strcmp(key, COMMAND_SIZE)) {
            // You don't need change the map size.
            ;
        } else if (!strcmp(key, COMMAND_PRINT)) {
            cout << "<contents of the AssociativeArray>" << endl;
            map<string, void*>::iterator pos;
            for (pos = _map.begin(); pos != _map.end(); pos++) {
                cout << "key: " << pos->first << ", value: " << (int) pos->second << endl;
            }
        } else {
            _map[string(key)] = element;
	      }
    }
    void* at(const char *key) {
        if (!strcmp(key, COMMAND_SIZE)) {
            return (void*) _map.size();
        } else {
            map<string, void*>::iterator pos;
            for (pos = _map.begin(); pos != _map.end(); pos++) {
                if (!strcmp(key, (pos->first).c_str()))
                    return (void *)pos->second;
            }
            return (void *)0;
        }
    }
private:
    map<string, void*> _map;
};

class AssociativeArrayEnum {
public:
    enum Key { K_print = 0,
               K_size,
               K_CircularList_example,
               K_AssociativeArray_example };

    AssociativeArrayEnum(int size=300): _position(K_size + 1), _size(size) {
        _pa   = new void*[_size];
    }

    ~AssociativeArrayEnum() {
        if (_pa) delete[] _pa;
    }

    void  atPut(void* element, const Key key) {
        if (key == K_print) {
            cout << "<contents of the AssociativeArray>" << endl;
            for (int i = K_size + 1; i < _position; i++) {
                cout << "key: " << i << ", value: " << (int) _pa[i] << endl;
            }
        } else if (key == K_size) {
            int i;
            void ** _tpa = new void*[(int) element];
            if (_pa) {
                for (i = K_size + 1; i < _position; i++) {
                    if ((int) element == i) {
                        break;
                    } else {
                        _tpa[i] = _pa[i];
                    }
                }
                _size = (int) element;
                delete [] _pa;
                _pa = _tpa;
	    }
        } else if (_pa && key >= 0 && key < _size) {
            _pa[key] = element;
            _position++;
        }
    }
    void* at(const Key key) {
        if (key == K_size) {
            return (void*) _size;
        } else if (_pa && key >=0 && key < _size)
            return _pa[key];
        else
            return (void *)0;
    }
private:
    int     _position;
    int     _size;
    void**  _pa;
};

int main() {
    int numIter = 1000000;

    AssociativeArrayMap am;
    clock_t start = clock();
    for (int i = 0; i < numIter; i++) {
        am.atPut((void *)230, "CircularList example");
        am.at("CircularList example");
    }
    clock_t end = clock();

    cout << "example 6-4 associative arrays takes " << (double)(end - start)/(double)CLK_TCK << " seconds." << endl;

    AssociativeArrayEnum ae(numIter + 2);
    start = clock();
    for (int j = 0; j < numIter; j++) {
        ae.atPut((void *)230, (AssociativeArrayEnum::Key) (AssociativeArrayEnum::K_CircularList_example + j));
        ae.at((AssociativeArrayEnum::Key) (AssociativeArrayEnum::K_CircularList_example + j));
    }
    end = clock();

    cout << "example 6-5 associative arrays takes " << (double)(end - start)/(double)CLK_TCK << " seconds." << endl;

    return 0;
}
