/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 9-11 */

#include <iostream.h>
#include <string.h>

// Top & Thing class definition
//
class Top {
public:
    virtual ~Top() {}
    virtual Top* type() { return this; }
    static void operator delete(void *p) { ::operator delete(p); }

protected:
    Top() {}
    static void* operator new(size_t l) { return ::operator new(l); }
};

typedef unsigned long REF_TYPE;
class Thing : public Top {
public:
    Thing(): refCountVal(1), updateCountVal(0) {}
    virtual ~Thing() {}
    
    virtual REF_TYPE deref() { return --refCountVal; }
    virtual REF_TYPE ref()   { return ++refCountVal; }
    virtual Thing* cutover(Thing*) { return this; }
    
private:
    REF_TYPE refCountVal, updateCountVal;
};

// StringRep class & its derived classes definition
//
class String;
class StringRep : public Thing {
public:
    StringRep()  { rep = new char[1]; rep[0] = '\0'; }
    ~StringRep() { delete rep; }
    virtual String make(const char* s="");
    virtual int    length() const { return strlen(rep); }
    virtual char*  c_str() const  { return rep; }

protected:
    friend class String;
    static void* operator new(size_t l)  { return ::operator new(l); }
    static void operator delete(void* p) { ::operator delete(p); }
    StringRep(const char* s) { rep = new char[strlen(s)+1]; strcpy(rep, s); }

private:
    char *rep;
};

static StringRep stringRepExemplar;
StringRep* stringRep = &stringRepExemplar;

// String class definition
//
class String : public Top {
public:
    String()              { rep = (stringRep->make()).rep; }
    String(const char* s) { rep = (stringRep->make(s)).rep; }
    String(const String& s)     { (rep = s.rep)->ref();  }
    String(StringRep& s)  { (rep = &s)->ref();  }
    ~String() { if (rep && rep->deref() <=0) delete rep; }

    String& operator=(const String& s) { 
        if (rep == s.rep) return *this;
        if (rep && rep->deref() <= 0) delete rep;
        (rep = s.rep)->ref();
        return *this;
    }
    StringRep* operator->() const { return rep; }    
private:
    static void* operator new(size_t) { cerr << "heap string" << endl; }
    static void  operator delete(void*) {}
    StringRep* rep;

};

static String stringExemplar;
String* string = &stringExemplar;

// StringRep class implementation
//
String StringRep::make(const char* s) {
    StringRep* rep = new StringRep(s);
    return String(*rep);
}

// Function for testing
//
int main() {
    String s1 = (*string)->make("sample string");
    cout << "string content = " << s1->c_str() << endl;
    cout << "string length  = " << s1->length() << endl << endl;

    String s2(s1);
    cout << "string content = " << s2->c_str() << endl;
    cout << "string length  = " << s2->length() << endl << endl;

    String s3 = s2;
    cout << "string content = " << s3->c_str() << endl;
    cout << "string length  = " << s3->length() << endl << endl;

    return 0;
}
