/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 9-9 */

#include <iostream.h>
#include <list>
#include <string.h>

using namespace std;

#define round(s) ((size_t)((s + sizeof(void*) - 1)/sizeof(void*))*sizeof(void*))

typedef char *Char_p;

static const int AGE_THRESHOLD = 1;

struct Coordinate {
    int x, y;
    Coordinate(int xx = 0, int yy = 0) { x = xx; y = yy; }
};

class Exemplar { public: Exemplar() {} };

// Top & Thing class definition
//
class Top {
public:
    virtual ~Top() {}
    virtual Top* type() { return this; }
    static void operator delete(void *p) { ::operator delete(p); }

protected:
    Top() {}
    static void* operator new(size_t l)  { return ::operator new(l); }
};
typedef Top* Topp;

class Thing : public Top {
public:
    Thing() {}
    virtual ~Thing() {}
};
typedef Thing* Thingp;

// StringRep class & its derived classes definition
//
class  Shape;
class ShapeRep : public Thing {
public:
    ShapeRep(Exemplar);
    ShapeRep()        { space = ToSpace; }
    ~ShapeRep()       { /* do nothing */ }

    Shape make(Coordinate, Coordinate, Thingp);
    Shape make(Coordinate, Coordinate, Coordinate, Thingp);

    virtual Shape make();
    virtual Shape make(Coordinate, Coordinate);
    virtual Shape make(Coordinate, Coordinate, Coordinate);

    virtual void rotate(double)   {}
    virtual void move(Coordinate) {}
    virtual void draw()           { cout << "<ShapeRep>" << endl; }
    virtual void erase()          {}

    Top*    type()                { return exemplarPointer; }
    virtual void mark()           { gcmark = 1; }
    virtual void gc(size_t=0)     { }
    virtual void *copy2(ShapeRep*) { }
    virtual void *copy3(ShapeRep*) { }
    static  void init();

protected:
    friend class Shape;
    Coordinate   center;
    Thing*       exemplarPointer;
    unsigned     gcmark:1, inUse:1, space:1;
    static unsigned char FromSpace, ToSpace;
    int age;

    static void  gcCommon(size_t, const size_t, const int, Char_p&, Char_p&, Char_p&);

    static void* operator new(size_t l)    { return ::operator new(l); }
    static void  operator delete(void* p)  { ::operator delete(p); }

    static Shape* aShape;
};

Shape* ShapeRep::aShape = 0;
unsigned char ShapeRep::FromSpace = 0;
unsigned char ShapeRep::ToSpace = 1;

class Triangle : public ShapeRep {
public:
    Triangle(Exemplar e) : ShapeRep(e) {}
    Triangle() { /* do nothing */ }
    Triangle(const Triangle& t) { p1 = t.p1; p2 = t.p2; }

    Shape make();
    Shape make(Coordinate c1, Coordinate c2, Coordinate c3);

    void rotate(double)   {}
    void move(Coordinate) {}
    void draw()           { cout << "<Triangle>"; }
    void erase()          {}

    void* operator new(size_t nbytes);
    void* copy2(ShapeRep*);
    void* copy3(ShapeRep*);
    void  operator delete(void*)      {}
    void  gc(size_t nbytes = 0)       { gcCommon(nbytes, poolInitialized, PoolSize, heap1, heap2, heap3); }
    static void init();

protected:
    Triangle(Coordinate c1, Coordinate c2, Coordinate c3) { p1 = c1; p2 = c2; p3 = c3; }

private:
    Coordinate p1, p2, p3;

    static char*    heap1; // for generation 1
    static char*    heap2; // for generation 2
    static char*    heap3; // for generation 3
    static size_t   poolInitialized;
    enum { PoolSize = 10 };
};

ShapeRep* triangle = 0;
char*  Triangle::heap1 = 0;
char*  Triangle::heap2 = 0;
char*  Triangle::heap3 = 0;
size_t Triangle::poolInitialized = 0;

class Rectangle : public ShapeRep {
public:
    Rectangle(Exemplar e) : ShapeRep(e) {}
    Rectangle()                   { /* do nothing */ }
    Rectangle(const Rectangle& r) { p1 = r.p1; p2 = r.p2; }

    Shape make();
    Shape make(Coordinate, Coordinate);

    void rotate(double)   {}
    void move(Coordinate) {}
    void draw()           { cout << "<Rectangle>"; }
    void erase()          {}

    void* operator new(size_t nbytes);
    void* copy2(ShapeRep*);
    void* copy3(ShapeRep*);
    void  operator delete(void*)      {}
    void  gc(size_t nbytes = 0)       { gcCommon(nbytes, poolInitialized, PoolSize, heap1, heap2, heap3); }
    static void init();

protected:
    Rectangle(Coordinate c1, Coordinate c2) { p1 = c1; p2 = c2; }

private:
    Coordinate p1, p2;

    static char*    heap1; // for generation 1
    static char*    heap2; // for generation 2
    static char*    heap3; // for generation 3
    static size_t   poolInitialized;
    enum { PoolSize = 10 };
};

ShapeRep* rectangle = 0;
char*  Rectangle::heap1 = 0;
char*  Rectangle::heap2 = 0;
char*  Rectangle::heap3 = 0;
size_t Rectangle::poolInitialized = 0;

// Shape class definition
//
class Shape : public Top {
public:
    Shape()              { allShapes->push_back((Topp)this); rep = new ShapeRep; }
    Shape(const Shape& s)      { allShapes->push_back((Topp)this); rep = s.rep; }
    Shape(ShapeRep& sr)  { allShapes->push_back((Topp)this); rep = &sr; }
    ~Shape()             { list<Topp>::iterator iter = allShapes->begin();
                           for (; iter != allShapes->end(); iter++)
                               if (*iter == (Topp)this) { allShapes->erase(iter); break; }
                           if (allShapes->empty()) gc(); }

    Shape& operator=(const Shape& s)   { rep = s.rep; return *this; }
    ShapeRep* operator->() const { return rep; }
    Top*   type()                { extern Shape* shape; return shape; }

    void Register(ShapeRep* s)   { allShapeExemplars->push_back((Thingp)s); }
    void UnRegister(ShapeRep* s) { list<Thingp>::iterator iter = allShapeExemplars->begin();
                                   for (; iter != allShapeExemplars->end(); iter++)
                                       if (*iter == (Thingp)s) { allShapeExemplars->erase(iter); break; } }
    void gc() {
        if (!gcmarked) {
            list<Topp>::iterator iter1 = allShapes->begin();
            for (; iter1 != allShapes->end(); iter1++) {
                Shape* sp = (Shape*)(*iter1);
                if (sp->rep) sp->rep->mark();
            }

            list<Thingp>::iterator iter2 = allShapeExemplars->begin();
            for (; iter2 != allShapeExemplars->end(); iter2++) {
                ShapeRep* exemplar = (ShapeRep*)(*iter2);
                exemplar->space = ShapeRep::FromSpace;
            }

            ShapeRep::FromSpace ^= 1;
            ShapeRep::ToSpace   ^= 1;
            gcmarked ^= 1;
        }
        else {
            int count = 1;
            list<Thingp>::iterator iter2 = allShapeExemplars->begin();
            for (; iter2 != allShapeExemplars->end(); ++iter2, count++) {
                ShapeRep* exemplar = (ShapeRep*)(*iter2);
                if (exemplar->space == ShapeRep::ToSpace ) {
                    exemplar->gc(0);
                    exemplar->space ^= 1;
                    // for testing, stop gc after sweeping one exemplar.
                    break;
                }
            }

            if (count == allShapeExemplars->size()) {
                ShapeRep::FromSpace ^= 1;
                ShapeRep::ToSpace   ^= 1;
                gcmarked ^= 1;
            }
        }
    }

    static void init() {
        allShapes = new list<Topp>;
        allShapeExemplars = new list<Thingp>;

        extern Shape* shape;
        shape = new Shape;

        ShapeRep::init();
        Triangle::init();
        Rectangle::init();
    }

private:
    friend class ShapeRep;
    static void* operator new(size_t s)  { return ::operator new(s); }
    static void  operator delete(void*)  { }

    static unsigned char gcmarked;
    static list<Topp>*   allShapes;
    static list<Thingp>* allShapeExemplars;

    ShapeRep* rep;
};

Shape* shape = 0;
unsigned char Shape::gcmarked  = 0;
list<Topp>*   Shape::allShapes = 0;
list<Thingp>* Shape::allShapeExemplars = 0;

// StringRep class & its derived classes implementation
//
ShapeRep::ShapeRep(Exemplar) {
    space = ToSpace;
    exemplarPointer = this;
    shape->Register(this);
}
Shape ShapeRep::make(Coordinate c1, Coordinate c2, Thingp type) {
    return (((ShapeRep*)type)->make(c1, c2));
}
Shape ShapeRep::make(Coordinate c1, Coordinate c2, Coordinate c3, Thingp type) {
    return (((ShapeRep*)type)->make(c1, c2, c3));
}
Shape ShapeRep::make() {
    return *aShape;
}
Shape ShapeRep::make(Coordinate c1, Coordinate c2) {
    return make(c1, c2, rectangle);
}
Shape ShapeRep::make(Coordinate c1, Coordinate c2, Coordinate c3) {
    return make(c1, c2, c3, triangle);
}
void ShapeRep::init() {
    aShape = new Shape;
}
void ShapeRep::gcCommon(size_t nbytes, const size_t poolInitialized, const int PoolSize, Char_p& heap1, Char_p& heap2, Char_p& heap3) {
    int i, j;
    // initialization
    if (!poolInitialized) {
        size_t    size = round(nbytes);
        heap1 = new char[PoolSize * size];
        heap2 = new char[PoolSize * size];
        heap3 = new char[PoolSize * size];

        ShapeRep*   tp = (ShapeRep*)heap1;
        for (i = 0; i < PoolSize; i++) {
            tp->inUse  = 0;
            tp->gcmark = 0;
            tp->age = 0;
            tp = (ShapeRep*)(Char_p(tp) + size);
        }
        tp = (ShapeRep*)heap2;
        for (i = 0; i < PoolSize; i++) {
            tp->inUse  = 0;
            tp->gcmark = 0;
            tp->age = 0;
            tp = (ShapeRep*)(Char_p(tp) + size);
        }
        tp = (ShapeRep*)heap3;
        for (i = 0; i < PoolSize; i++) {
            tp->inUse  = 0;
            tp->gcmark = 0;
            tp->age = 0;
            tp = (ShapeRep*)(Char_p(tp) + size);
        }
    }
    // normal garbage collection
    else {
        size_t    size = round(poolInitialized);
        ShapeRep*   tp = (ShapeRep*)heap1;
        for (i = 0; i < PoolSize; i++, tp = (ShapeRep*)(Char_p(tp) + size)) {
            tp->age++;
            if (tp->age == AGE_THRESHOLD && tp->inUse) {
                cout << "object moved to intermediate pool" << endl;
                tp->copy2(tp);
                tp->inUse  = 0;
                tp->gcmark = 0;
                tp->age = 0;
            }
            // no need to reclaim if block is not in use
            if (tp->inUse == 0)   continue;

            // should not reclaim if exemplar
            if (tp == tp->type()) continue;

            // should not reclaim if block is marked
            if (tp->gcmark)  { tp->gcmark = 0; tp->space  = FromSpace; continue; }

            // should not reclaim if block is allocated after mark phase
            if (tp->space == ToSpace) { tp->space  = FromSpace; continue; }

            // reclaim the block now.
            cout << "ShapeRep::gcCommon - reclaimed object ";
            tp->draw(); cout << " " << i << endl;

            tp->ShapeRep::~ShapeRep();
            tp->inUse  = 0;
        }
        tp = (ShapeRep*)heap2;
        for (i = 0; i < PoolSize; i++, tp = (ShapeRep*)(Char_p(tp) + size)) {
            tp->age++;
            if (tp->age == AGE_THRESHOLD && tp->inUse) {
                cout << "object moved to old pool" << endl;
                tp->copy3(tp);
                tp->inUse  = 0;
                tp->gcmark = 0;
                tp->age = 0;
            }
            // no need to reclaim if block is not in use
            if (tp->inUse == 0)   continue;

            // should not reclaim if exemplar
            if (tp == tp->type()) continue;

            // should not reclaim if block is marked
            if (tp->gcmark)  { tp->gcmark = 0; tp->space  = FromSpace; continue; }

            // should not reclaim if block is allocated after mark phase
            if (tp->space == ToSpace) { tp->space  = FromSpace; continue; }

            // reclaim the block now.
            cout << "ShapeRep::gcCommon - reclaimed object ";
            tp->draw(); cout << " " << i << endl;

            tp->ShapeRep::~ShapeRep();
            tp->inUse  = 0;
        }
        tp = (ShapeRep*)heap3;
        for (i = 0; i < PoolSize; i++, tp = (ShapeRep*)(Char_p(tp) + size)) {
            tp->age++;
            // no need to reclaim if block is not in use
            if (tp->inUse == 0)   continue;

            // should not reclaim if exemplar
            if (tp == tp->type()) continue;

            // should not reclaim if block is marked
            if (tp->gcmark)  { tp->gcmark = 0; tp->space  = FromSpace; continue; }

            // should not reclaim if block is allocated after mark phase
            if (tp->space == ToSpace) { tp->space  = FromSpace; continue; }

            // reclaim the block now.
            cout << "ShapeRep::gcCommon - reclaimed object ";
            tp->draw(); cout << " " << i << endl;

            tp->ShapeRep::~ShapeRep();
            tp->inUse  = 0;
        }
    }
}

Shape Triangle::make() {
    Coordinate c(0,0);
    Triangle* tri = new Triangle(c, c, c);
    return Shape(*tri);
}
Shape Triangle::make(Coordinate c1, Coordinate c2, Coordinate c3) {
    Triangle* tri = new Triangle(c1, c2, c3);
    return Shape(*tri);
}
void Triangle::init() {
    triangle = new Triangle(Exemplar());
}
void* Triangle::operator new(size_t nbytes) {
    int i;
    if (!poolInitialized) {
        gcCommon(nbytes, poolInitialized, PoolSize, heap1, heap2, heap3);
        poolInitialized = nbytes;
    }

    Triangle* p = (Triangle*)heap1;
    for (i = 0; i < PoolSize; i++) {
        if (p->inUse == 0) break;
        p = (Triangle*)(Char_p(p) + round(nbytes));
    }
    if (i == PoolSize) throw;

    p->age = 0;
    p->inUse  = 1;
    p->gcmark = 0;
    return (void*)p;
}

void* Triangle::copy2(ShapeRep* src) {
    int i;

    Triangle* p = (Triangle*)heap2;
    for (i = 0; i < PoolSize; i++) {
        if (p->inUse == 0) break;
        p = (Triangle*)(Char_p(p) + round(sizeof(Triangle)));
    }
    if (i == PoolSize) throw;
    ::memcpy(p, src, sizeof(Triangle));
    p->age = 0;
    p->inUse  = 1;
    p->gcmark = 0;
    return (void*)p;
}

void* Triangle::copy3(ShapeRep* src) {
    int i;

    Triangle* p = (Triangle*)heap3;
    for (i = 0; i < PoolSize; i++) {
        if (p->inUse == 0) break;
        p = (Triangle*)(Char_p(p) + round(sizeof(Triangle)));
    }
    if (i == PoolSize) throw;
    ::memcpy(p, src, sizeof(Triangle));
    p->age = 0;
    p->inUse  = 1;
    p->gcmark = 0;
    return (void*)p;
}

Shape Rectangle::make() {
    Coordinate c(0,0);
    Rectangle* rect = new Rectangle(c, c);
    return Shape(*rect);
}
Shape Rectangle::make(Coordinate c1, Coordinate c2) {
    Rectangle* rect = new Rectangle(c1, c2);
    return Shape(*rect);
}
void Rectangle::init() {
    rectangle = new Rectangle(Exemplar());
}
void* Rectangle::operator new(size_t nbytes) {
    int i;
    if (!poolInitialized) {
        gcCommon(nbytes, poolInitialized, PoolSize, heap1, heap2, heap3);
        poolInitialized = nbytes;
    }

    Rectangle* p = (Rectangle*)heap1;
    for (i = 0; i < PoolSize; i++) {
        if (p->inUse == 0) break;
        p = (Rectangle*)(Char_p(p) + round(nbytes));
    }
    if (i == PoolSize) throw;

    p->age  = 0;
    p->inUse  = 1;
    p->gcmark = 0;
    return (void*)p;
}

void* Rectangle::copy2(ShapeRep* src) {
    int i;

    Rectangle* p = (Rectangle*)heap2;
    for (i = 0; i < PoolSize; i++) {
        if (p->inUse == 0) break;
        p = (Rectangle*)(Char_p(p) + round(sizeof(Rectangle)));
    }
    if (i == PoolSize) throw;
    ::memcpy(p, src, sizeof(Rectangle));
    p->age = 0;
    p->inUse  = 1;
    p->gcmark = 0;
    return (void*)p;
}

void* Rectangle::copy3(ShapeRep* src) {
    int i;

    Rectangle* p = (Rectangle*)heap3;
    for (i = 0; i < PoolSize; i++) {
        if (p->inUse == 0) break;
        p = (Rectangle*)(Char_p(p) + round(sizeof(Rectangle)));
    }
    if (i == PoolSize) throw;
    ::memcpy(p, src, sizeof(Rectangle));
    p->age = 0;
    p->inUse  = 1;
    p->gcmark = 0;
    return (void*)p;
}

// Function for testing
//
int main() {
    Shape::init();

    Coordinate c1(1,1), c2(2,2), c3(3,3);

    Shape t1 = (*shape)->make(c1, c2, c3);
    t1 = (*shape)->make(c1, c2, c3);

    shape->gc();      // mark phase

    t1 = (*shape)->make(c1, c2, c3);

    shape->gc();     // sweep phase should reclaim one triangle object.

    Shape r1 = (*shape)->make(c1, c2);
    r1 = (*shape)->make(c1, c2);

    shape->gc();     // sweep the rectangle exemplar.

    r1 = (*shape)->make(c2, c3);

    shape->gc();     // mark phase
    shape->gc();     // sweep phase should reclaim one triangle object.
    shape->gc();     // sweep phase should reclaim two rectangle objects.

    return 0;
}
